#!/usr/bin/python3

# Use fast astrometry on a CCD image
# Update guiding image center coordinate and rotation files 
#   from the fit to the image field

# Copyright (c) 2017 John Kielkopf and Rhodes Hart        
# kielkopf@louisville.edu                                                  
#                                                                          
#                                                                          
# Date: September 25, 2017                                                
# Version: 1.1                                                          
#                                                                          
# This file is part of XmCCD and XmTel for the CDK20                                             

# History 
#   2017-05-19  1.0 First code optimized for CDK20N
#   2017-07-28  1.1 Modified for CDK700
#   2017-09-24  1.2 Updated status file names and for CDK20

# Note image scale hints are hard coded for a specific telescope and camera

import sys
import os
from subprocess import Popen, PIPE
import string
import time
import math
import re
import glob

if len(sys.argv) == 1:
  telcoordsfp = open("/usr/local/observatory/status/telcoords", 'r')
  telcoordstext = telcoordsfp.readlines()
  for line in telcoordstext:
    try:
      entry = line.strip().split()
      rahr = float(entry[0])
      radeg = 15.*rahr
      decdeg = float(entry[1]) 
    except:
      radeg = 0.
      decdeg = 0.
else:
  print (" ")
  print ("Reads system /usr/local/observatory/status/telcoords for a hint.")
  print ("Tries fast astrometry on the most recent fits file in current directory.")
  print ("Updates the guidecenter and guiderotation status files.")
  sys.exit ("No argument is required.")
  

# Talk to astrometry and find coordinates
# Requires temporary working file directory /home/observer/data/tmp

file_name = max(glob.iglob('*.fits'), key=os.path.getctime)

print ("Running astrometry on ", file_name, " seeded from telcoords status file")

astrometry = "/usr/local/astrometry/bin/solve-field  --no-plots  -z 2" \
  " --dir /home/john/data/tmp/" \
  " --scale-units arcsecperpix --scale-low 0.52 --scale-high 0.55" \
  " --radius 6.0 --ra %f --dec %f"  \
  " --overwrite --new-fits none  %s" % (radeg, decdeg, file_name)


p1 = Popen( astrometry , shell=True, stdout=PIPE)
p1.wait()
output = str(p1.communicate()[0], 'utf-8')

# Parse stdout in phases

scratch, center = output.split("Dec D:M:S) = (",1)

ratxt = center.split()[0]
dectxt = center.split()[1]
rotationtxt = center.split()[13]

rahr =  re.findall(r"[-+]?\d*\.\d+|\d+",ratxt.split(":")[0])
ramin = re.findall(r"[-+]?\d*\.\d+|\d+",ratxt.split(":")[1])
rasec = re.findall(r"[-+]?\d*\.\d+|\d+",ratxt.split(":")[2])

decdeg = re.findall(r"[-+]?\d*\.\d+|\d+",dectxt.split(":")[0])
decmin =  re.findall(r"[-+]?\d*\.\d+|\d+",dectxt.split(":")[1])
decsec =  re.findall(r"[-+]?\d*\.\d+|\d+",dectxt.split(":")[2])

rotation = re.findall(r"[-+]?\d*\.\d+|\d+",rotationtxt)

decsign = 1.
if ( float(decdeg[0]) < 0. ):
  decsign = -1.
  
ra = (float(rahr[0]) + float(ramin[0])/60. + float(rasec[0])/3600.)
dec = decsign*(abs(float(decdeg[0])) + float(decmin[0])/60. + float(decsec[0])/3600.)
rot = float(rotation[0])

# Update status of guiding field center 

guidecenterfp = open("/usr/local/observatory/status/guidecoords", 'w')
guidecentertext = "%12.6f  %12.6f " % (ra, dec) 
guidecenterfp.write(guidecentertext)
guidecenterfp.close()

# Update status of guiding field rotation

guiderotationfp = open("/usr/local/observatory/status/guiderotate", 'w')
guiderotationtext = "%5.3f" % (rot) 
guiderotationfp.write(guiderotationtext)
guiderotationfp.close()

print ("\n")
print ("Image: ", file_name)
print ("RA [hr]: %10.5f" % ra)
print ("Declination [deg]: %10.5f" % dec)
print ("Field rotation [deg]: %5.3f" % rot)
print ("Written to guidecoords and guiderotate status files")
print ("\n")

exit()
  
